import datetime
import logging
import re

import numpy as np
import pandas as pd

from wbportfolio.import_export.utils import convert_string_to_number
from wbportfolio.models import Fees

logger = logging.getLogger("importers.parsers.jpmorgan.fee")


def file_name_parse(file_name):
    dates = re.findall("([0-9]{8})", file_name)

    assert len(dates) == 1, "Not exactly 1 date found in the filename"

    return {"valuation_date": datetime.datetime.strptime(dates[0], "%Y%m%d").date()}


def parse(import_source):
    df = pd.read_csv(import_source.file, encoding="utf-16", delimiter=",")
    df["Date"] = pd.to_datetime(df["Date"])

    columns_fees = df.columns.difference(["ISIN", "CCY"])
    df[columns_fees] = df[columns_fees].where(pd.notnull(df[columns_fees]), 0)
    df = df.replace([np.inf, -np.inf, np.nan], None)

    # Iterate through the CSV File and parse the data into a list
    data = list()

    for fee_data in df.to_dict("records"):
        isin = fee_data["ISIN"]
        fee_date = fee_data["Date"]
        base_data = {
            "product": {"isin": isin},
            "fee_date": fee_date.strftime("%Y-%m-%d"),
            "calculated": False,
        }
        data.append(
            {
                "transaction_subtype": Fees.Type.PERFORMANCE,
                "total_value": round(convert_string_to_number(fee_data.get("Perf_Fee * Units", 0)), 4),
                "total_value_gross": round(convert_string_to_number(fee_data.get("Perf_Fee * Units", 0)), 4),
                **base_data,
            }
        )
        data.append(
            {
                "transaction_subtype": Fees.Type.MANAGEMENT,
                "total_value": round(convert_string_to_number(fee_data.get("Mgmt_Fee * Units", 0)), 4),
                **base_data,
            }
        )
        data.append(
            {
                "transaction_subtype": Fees.Type.ISSUER,
                "total_value": round(convert_string_to_number(fee_data.get("JPM_Fee * Units", 0)), 4),
                **base_data,
            }
        )

    return {"data": data}



# Generated by Django 4.1.9 on 2023-07-11 08:24

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("io", "0001_initial_squashed"),
        ("wbportfolio", "0041_remove_listresearch_research_ptr_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="InstrumentList",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255)),
                ("identifier", models.CharField(max_length=255, unique=True)),
                (
                    "instrument_list_type",
                    models.CharField(
                        blank=True,
                        choices=[("WATCH", "Watch List"), ("EXCLUSION", "Exclusion List")],
                        max_length=32,
                        null=True,
                    ),
                ),
            ],
            options={
                "verbose_name": "Instrument List",
                "verbose_name_plural": "Instrument Lists",
            },
        ),
        migrations.CreateModel(
            name="InstrumentListThroughModel",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("from_date", models.DateField(blank=True, null=True)),
                ("to_date", models.DateField(blank=True, null=True)),
                ("comment", models.TextField(blank=True, default="")),
                ("validated", models.BooleanField(default=False)),
                (
                    "import_source",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="io.importsource"
                    ),
                ),
                (
                    "instrument",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="wbportfolio.instrument"),
                ),
                (
                    "instrument_list",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="wbportfolio.instrumentlist"),
                ),
            ],
            options={
                "verbose_name": "Instruments in Instrument List",
                "unique_together": {("instrument", "instrument_list")},
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name="instrumentlist",
            name="instruments",
            field=models.ManyToManyField(
                through="wbportfolio.InstrumentListThroughModel", to="wbportfolio.instrument"
            ),
        ),
    ]



from wbcore.metadata.configs.endpoints import EndpointViewConfig

from wbportfolio.models.reconciliations.account_reconciliations import (
    AccountReconciliation,
)


class AccountReconciliationEndpointViewConfig(EndpointViewConfig):
    def get_instance_endpoint(self, **kwargs):
        if self.view.kwargs.get("pk"):
            return None
        return super().get_instance_endpoint(**kwargs)

    def get_delete_endpoint(self, **kwargs):
        return None


class AccountReconciliationLineEndpointViewConfig(EndpointViewConfig):
    def get_instance_endpoint(self, **kwargs):
        return None

    def get_delete_endpoint(self, **kwargs):
        return None

    def get_create_endpoint(self, **kwargs):
        return None

    def get_update_endpoint(self, **kwargs):
        if accountreconciliation_id := self.view.kwargs.get("accountreconciliation_id"):
            if AccountReconciliation.objects.filter(id=accountreconciliation_id, approved_dt__isnull=False).exists():
                return None
        return super().get_endpoint()



from wbcore.metadata.configs.titles import TitleViewConfig

from wbportfolio.models import Product


class InstrumentPriceAUMDataTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Assets under Management"


# Product Portfolio Viewsets
class NominalProductTitleConfig(TitleViewConfig):
    def get_list_title(self):
        product = Product.objects.get(id=self.view.kwargs["product_id"])
        return f"{str(product)}: Nominal Value"


class AUMProductTitleConfig(TitleViewConfig):
    def get_list_title(self):
        product = Product.objects.get(id=self.view.kwargs["product_id"])
        return f"{str(product)}: Assets under Management"



from typing import Generator

from django.contrib.contenttypes.models import ContentType
from wbcompliance.models.risk_management import backend
from wbcompliance.models.risk_management.dispatch import register
from wbcore import serializers as wb_serializers
from wbcore.serializers.fields.number import percent_decorator
from wbfdm.models import Instrument

from wbportfolio.pms.typing import Portfolio as PortfolioDTO

from .mixins import ActiveProductRelationshipMixin


@register("Liquidity Stress Instrument Rule Backend", rule_group_key="portfolio")
class RuleBackend(ActiveProductRelationshipMixin, backend.AbstractRuleBackend):
    @classmethod
    def get_parameter_fields(cls):
        return [
            "group_by",
            "field",
            "is_cash",
            "asset_classes",
            "currencies",
            "countries",
            "classification_height",
            "classifications",
        ]

    @classmethod
    def get_serializer_class(cls) -> wb_serializers.Serializer:
        class RuleBackendSerializer(wb_serializers.Serializer):
            below_x_days = wb_serializers.IntegerField(default=5, label="Below X days")
            liquidity_factor = wb_serializers.FloatField(default=0.33, label="Liquidity factory")
            pct_worst_volume = wb_serializers.FloatField(
                default=100.0, label="Worst volume (%)", percent=True, decorators=[percent_decorator]
            )
            pct_redemption = wb_serializers.FloatField(
                default=100.0, label="Redemption (%)", percent=True, decorators=[percent_decorator]
            )
            last_x_trading_dates = wb_serializers.IntegerField(default=60, label="Number of trading days")
            is_slicing = wb_serializers.BooleanField(default=True, label="Slicing")

            @classmethod
            def get_parameter_fields(cls):
                return [
                    "below_x_days",
                    "liquidity_factor",
                    "pct_worst_volume",
                    "pct_redemption",
                    "last_x_trading_dates",
                    "is_slicing",
                ]

        return RuleBackendSerializer

    def _process_dto(self, portfolio: PortfolioDTO, **kwargs) -> Generator[backend.IncidentResult, None, None]:
        # TODO adapt to DTO framework
        factor = self.instrument.pct_liquidated_below_n_days(
            self.evaluation_date,
            below_n_days=self.below_x_days,
            liq_factor=self.liquidity_factor,
            pct_worst_volume=self.pct_worst_volume,
            pct_redemption=self.pct_redemption,
            last_x_trading_dates=self.last_x_trading_dates,
            is_slicing=self.is_slicing,
        )
        for threshold in self.thresholds:
            if factor is not None and threshold.is_inrange(factor):
                yield backend.IncidentResult(
                    breached_object=self.instrument,
                    breached_object_repr=str(self.instrument),
                    breached_value=str(factor),
                    report_details=dict(),
                    severity=threshold.severity,
                )

    @classmethod
    def get_allowed_content_type(cls) -> "ContentType":
        return ContentType.objects.get_for_model(Instrument)

    def is_passive_evaluation_valid(self) -> bool:
        if not self.instrument.portfolio:
            return False
        return (self.instrument.portfolio.imported_assets.filter(date=self.evaluation_date).exists()) and (
            self.instrument.valuations.filter(date=self.evaluation_date).exists()
        )



from datetime import timedelta
from decimal import Decimal

import pytest
from faker import Faker
from pandas._libs.tslibs.offsets import BDay

from wbportfolio.models import Product, Trade

fake = Faker()


@pytest.mark.django_db
class TestTradeModel:
    def test_init(self, trade):
        assert trade.id is not None

    def test_str(self, product, trade_factory):
        trade = trade_factory.create(underlying_instrument=product)
        assert str(trade) == f"{trade.product.ticker}:{trade.shares} ({trade.bank})"

    def test_save_with_price(self, trade_factory):
        trade = trade_factory(price=150)
        assert trade.price == 150

    def test_subquery_net_money(self, trade_factory, instrument_price_factory, product):
        v = instrument_price_factory.create(instrument=product)
        c2 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=v.date - timedelta(days=1),
            shares=-1,
            transaction_subtype=Trade.Type.REDEMPTION,
        )
        c1 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=v.date - timedelta(days=1),
            shares=1,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        trade_factory.create(  # noise
            underlying_instrument=product,
            transaction_date=v.date,
            shares=1,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        products_pos = Product.objects.annotate(value=Trade.subquery_net_money(v.date, v.date))
        assert products_pos.first().value == pytest.approx(c2.shares * c2.price + c1.shares * c1.price)

        products_pos = Product.objects.annotate(value=Trade.subquery_net_money(v.date, v.date, only_positive=True))
        assert products_pos.first().value == pytest.approx(c1.shares * c1.price)

        products_neg = Product.objects.annotate(value=Trade.subquery_net_money(v.date, v.date, only_negative=True))
        assert products_neg.first().value == pytest.approx(c2.shares * c2.price)

    @pytest.mark.parametrize("date", [(fake.date_object())])
    def test_subquery_shares_per_product(self, trade_factory, product, date):
        a1 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date - timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        trade_factory.create(  # noise
            underlying_instrument=product, transaction_date=date, transaction_subtype=Trade.Type.SUBSCRIPTION
        )
        qs = Product.objects.annotate(total_shares=Trade.subquery_shares_per_underlying_instrument(date))
        assert qs.first().total_shares == a1.shares

    @pytest.mark.parametrize("date", [(fake.date_object())])
    def test_subquery_shares_per_product_per_date(self, trade_factory, product, date):
        a1 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date - timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        trade_factory.create(  # noise
            underlying_instrument=product, transaction_date=date, transaction_subtype=Trade.Type.SUBSCRIPTION
        )
        qs = Product.objects.annotate(total_shares=Trade.subquery_shares_per_underlying_instrument(date))
        assert qs.first().total_shares == a1.shares

    @pytest.mark.parametrize("date", [(fake.date_object())])
    def test_subquery_net_new_money(self, trade_factory, product, date):
        a1 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date - timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        trade_factory.create(  # noise
            underlying_instrument=product,
            transaction_date=date,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        qs = Product.objects.annotate(total_shares=Trade.subquery_shares_per_underlying_instrument(date))
        assert qs.first().total_shares == a1.shares

    def test_claim_update_claimed_shares(self, customer_trade_factory, claim_factory):
        customer_trade = customer_trade_factory.create()
        claim_factory.create(status="PENDING", trade=customer_trade)
        customer_trade.refresh_from_db()
        assert customer_trade.claimed_shares == 0
        c2 = claim_factory.create(status="APPROVED", trade=customer_trade)
        customer_trade.refresh_from_db()
        assert customer_trade.claimed_shares == c2.shares


@pytest.mark.django_db
class TestTradeProposalModel:
    def test_init(self, trade_proposal):
        assert trade_proposal.id is not None


@pytest.mark.django_db
class TestTradeInstrumentPrice:
    def test_shares(self, portfolio, product_factory, trade_factory, instrument_price_factory):
        product = product_factory.create()
        price = instrument_price_factory.create(instrument=product)
        trade1 = trade_factory.create(
            portfolio=product.primary_portfolio,
            underlying_instrument=product,
            transaction_date=price.date - timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        trade2 = trade_factory.create(
            portfolio=product.primary_portfolio,
            underlying_instrument=product,
            transaction_date=price.date - timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        trade_factory.create(  # noise
            portfolio=product.primary_portfolio,
            underlying_instrument=product,
            transaction_date=price.date,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        assert price.shares == pytest.approx(trade1.shares + trade2.shares)

    def test_nominal_value(self, portfolio, product_factory, trade_factory, instrument_price_factory):
        product = product_factory.create()
        price = instrument_price_factory.create(instrument=product)
        trade1 = trade_factory.create(
            portfolio=product.primary_portfolio,
            underlying_instrument=product,
            transaction_date=price.date - timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        trade2 = trade_factory.create(
            portfolio=product.primary_portfolio,
            underlying_instrument=product,
            transaction_date=price.date - timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        trade_factory.create(  # noise
            portfolio=product.primary_portfolio,
            underlying_instrument=product,
            transaction_date=price.date,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        assert price.nominal_value == pytest.approx(Decimal(product.share_price * (trade1.shares + trade2.shares)))

    def test_subscription_link_to_internal_trade(self, weekday, trade_factory, product):
        shares = 100
        trade_factory.create(shares=shares, transaction_subtype=Trade.Type.BUY, transaction_date=weekday)  # noise 1
        trade_factory.create(
            shares=shares,
            transaction_subtype=Trade.Type.BUY,
            transaction_date=weekday + timedelta(days=Trade.TRADE_WINDOW_INTERVAL + 1),
            underlying_instrument=product,
        )  # noise 2
        subscription_trade1 = trade_factory.create(
            shares=shares,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            transaction_date=weekday,
            underlying_instrument=product,
        )  # noise 3
        internal_trade = trade_factory.create(
            shares=shares, transaction_subtype=Trade.Type.BUY, transaction_date=weekday, underlying_instrument=product
        )
        subscription_trade2 = trade_factory.create(
            shares=shares,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            transaction_date=weekday,
            underlying_instrument=product,
        )

        subscription_trade1.link_to_internal_trade()
        assert subscription_trade1.internal_trade == internal_trade
        assert subscription_trade1.marked_as_internal is True

        # Reset attribute to test the other side
        subscription_trade1.internal_trade = None
        subscription_trade1.marked_as_internal = False
        subscription_trade1.save()

        # Check that linking from the other site but with two similar subscriptions won't link
        internal_trade.link_to_internal_trade()
        subscription_trade1.refresh_from_db()
        subscription_trade2.refresh_from_db()
        assert subscription_trade1.internal_trade is None
        assert subscription_trade2.internal_trade is None

        # Check that with now only one subscription, the linking is successful
        subscription_trade2.delete()
        internal_trade.link_to_internal_trade()
        subscription_trade1.refresh_from_db()
        assert subscription_trade1.internal_trade == internal_trade
        assert subscription_trade1.marked_as_internal is True

    def test_last_underlying_quote_price(self, weekday, trade_factory, instrument_price_factory):
        trade = trade_factory.create(transaction_date=weekday, value_date=(weekday - BDay(1)).date())
        assert trade.last_underlying_quote_price is None
        del trade.last_underlying_quote_price

        # test that underlying quote price returns any price found at transaction_date, or then at value_date (in that order)
        p0 = instrument_price_factory.create(instrument=trade.underlying_instrument, date=trade.value_date)
        assert trade.last_underlying_quote_price == p0
        del trade.last_underlying_quote_price

        p1 = instrument_price_factory.create(instrument=trade.underlying_instrument, date=trade.transaction_date)
        assert trade.last_underlying_quote_price == p1



import factory

from wbportfolio.factories.portfolios import PortfolioFactory
from wbportfolio.models import PortfolioSwingPricing


class PortfolioSwingPricingFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = PortfolioSwingPricing

    valid_date = factory.Faker("date_object")
    portfolio = factory.SubFactory(PortfolioFactory)
    negative_threshold = factory.Faker("pydecimal", left_digits=0, right_digits=4, min_value=-0.99, max_value=-0.01)
    negative_swing_factor = factory.Faker("pydecimal", left_digits=0, right_digits=4, min_value=-0.99, max_value=-0.01)
    positive_threshold = factory.Faker("pydecimal", left_digits=0, right_digits=4, min_value=0.01, max_value=0.99)
    positive_swing_factor = factory.Faker("pydecimal", left_digits=0, right_digits=4, min_value=0.01, max_value=0.99)



import codecs
import csv
import datetime
import re

from wbportfolio.import_export.utils import convert_string_to_number


def file_name_parse(file_name):
    dates = re.findall(r"([0-9]{4}-[0-9]{2}-[0-9]{2})", file_name)
    isin = re.findall(r"\.([a-zA-Z0-9]*)_", file_name)

    assert len(dates) == 2, "Not 2 dates found in the filename"
    assert len(isin) == 1, "Not exactly 1 isin found in the filename"

    return {
        "isin": isin[0],
        "valuation_date": datetime.datetime.strptime(dates[0], "%Y-%m-%d").date(),
        "generation_date": datetime.datetime.strptime(dates[1], "%Y-%m-%d").date(),
    }


def parse(import_source):
    # Load file into a CSV DictReader
    csv_file = import_source.file.open()
    # csv_file_type = type(csv_file.read())

    # # If the file is a byte string, then we need to convert it.
    # if csv_file_type is bytes:
    csv_file = codecs.iterdecode(csv_file, "latin1")

    # # Read file into a CSV Dict Reader
    csv_reader = csv.DictReader(csv_file, delimiter=",")

    # # Iterate through the CSV File and parse the data into a list
    data = list()

    for valuation in csv_reader:
        valuation_date = datetime.datetime.strptime(valuation["NAV Date"], "%Y/%m/%d").date()
        isin = valuation["ISIN Code"]
        if valuation_date.weekday() not in [5, 6] and valuation["ShareCurrency"] == valuation["Ccy"]:
            data.append(
                {
                    "instrument": {"instrument_type": "product", "isin": isin},
                    "date": valuation_date.strftime("%Y-%m-%d"),
                    "net_value": round(convert_string_to_number(valuation["NAV per share"]), 6),
                    "calculated": False,
                }
            )

    csv_file.close()
    return {"data": data}



from wbcore import serializers as wb_serializers
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display import create_simple_display


class NormalizeSerializer(wb_serializers.Serializer):
    total_cash_weight = wb_serializers.FloatField(default=0, precision=4, percent=True)


class TradeProposalButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.DropDownButton(
                label="Tools",
                buttons=(
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:tradeproposal",),
                        key="replay",
                        icon=WBIcon.SYNCHRONIZE.icon,
                        label="Replay Trades",
                        description_fields="""
                        <p>Replay Trades. It will recompute all assets positions until next trade proposal day (or today otherwise) </p>
                        """,
                        action_label="Replay Trade",
                        title="Replay Trade",
                    ),
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:tradeproposal",),
                        key="reset",
                        icon=WBIcon.REGENERATE.icon,
                        label="Reset Trades",
                        description_fields="""
                            <p>Delete and recreate initial trades to from its associated model portfolio</p>
                            """,
                        action_label="Reset Trades",
                        title="Reset Trades",
                    ),
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:tradeproposal",),
                        key="normalize",
                        icon=WBIcon.EDIT.icon,
                        label="Normalize Trades",
                        description_fields="""
                            <p>Make sure all trades normalize to a total target weight of (100 - {{total_cash_weight}})%</p>
                            """,
                        action_label="Normalize Trades",
                        title="Normalize Trades",
                        serializer=NormalizeSerializer,
                        instance_display=create_simple_display([["total_cash_weight"]]),
                    ),
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:tradeproposal",),
                        key="deleteall",
                        icon=WBIcon.DELETE.icon,
                        label="Delete All Trades",
                        description_fields="""
                    <p>Delete all trades from this trade proposal?</p>
                    """,
                        action_label="Delete All Trades",
                        title="Delete All Trades",
                    ),
                ),
            ),
        }

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()



import logging
import math
from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal
from typing import Any, TypeVar

from celery import shared_task
from django.core.exceptions import ValidationError
from django.db import DatabaseError, models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from pandas._libs.tslibs.offsets import BDay
from wbcompliance.models.risk_management.mixins import RiskCheckMixin
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton
from wbcore.models import WBModel
from wbcore.utils.models import CloneMixin
from wbfdm.models import InstrumentPrice
from wbfdm.models.instruments.instruments import Cash, Instrument

from wbportfolio.models.roles import PortfolioRole
from wbportfolio.pms.trading import TradingService
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.pms.typing import Position as PositionDTO

from ..asset import AssetPosition, AssetPositionIterator
from ..exceptions import InvalidAnalyticPortfolio
from .trades import Trade

logger = logging.getLogger("pms")

SelfTradeProposal = TypeVar("SelfTradeProposal", bound="TradeProposal")


class TradeProposal(CloneMixin, RiskCheckMixin, WBModel):
    trade_date = models.DateField(verbose_name="Trading Date")

    class Status(models.TextChoices):
        DRAFT = "DRAFT", "Draft"
        SUBMIT = "SUBMIT", "Submit"
        APPROVED = "APPROVED", "Approved"
        DENIED = "DENIED", "Denied"
        FAILED = "FAILED", "Failed"

    comment = models.TextField(default="", verbose_name="Trade Comment", blank=True)
    status = FSMField(default=Status.DRAFT, choices=Status.choices, verbose_name="Status")
    rebalancing_model = models.ForeignKey(
        "wbportfolio.RebalancingModel",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="trade_proposals",
        verbose_name="Rebalancing Model",
        help_text="Rebalancing Model that generates the target portfolio",
    )
    portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", related_name="trade_proposals", on_delete=models.PROTECT, verbose_name="Portfolio"
    )
    creator = models.ForeignKey(
        "directory.Person",
        blank=True,
        null=True,
        related_name="trade_proposals",
        on_delete=models.PROTECT,
        verbose_name="Owner",
    )

    class Meta:
        verbose_name = "Trade Proposal"
        verbose_name_plural = "Trade Proposals"
        constraints = [
            models.UniqueConstraint(
                fields=["portfolio", "trade_date"],
                name="unique_trade_proposal",
            ),
        ]

    def save(self, *args, **kwargs):
        if not self.trade_date and self.portfolio.assets.exists():
            self.trade_date = (self.portfolio.assets.latest("date").date + BDay(1)).date()

        # if a trade proposal is created before the existing earliest trade proposal, we automatically shift the linked instruments inception date to allow automatic NAV computation since the new inception date
        if not self.portfolio.trade_proposals.filter(trade_date__lt=self.trade_date).exists():
            new_inception_date = (self.trade_date + BDay(1)).date()
            self.portfolio.instruments.filter(inception_date__gt=new_inception_date).update(
                inception_date=new_inception_date
            )
        super().save(*args, **kwargs)

    @property
    def check_evaluation_date(self):
        return self.trade_date

    @property
    def checked_object(self) -> Any:
        return self.portfolio

    @cached_property
    def portfolio_total_asset_value(self) -> Decimal:
        return self.portfolio.get_total_asset_value(self.last_effective_date)

    @cached_property
    def validated_trading_service(self) -> TradingService:
        """
        This property holds the validated trading services and cache it.This property expect to be set only if is_valid return True
        """
        target_portfolio = self.convert_to_portfolio()

        return TradingService(
            self.trade_date,
            effective_portfolio=self._get_default_effective_portfolio(),
            target_portfolio=target_portfolio,
            total_target_weight=target_portfolio.total_weight,
        )

    @cached_property
    def last_effective_date(self) -> date:
        try:
            return self.portfolio.assets.filter(date__lt=self.trade_date).latest("date").date
        except AssetPosition.DoesNotExist:
            return self.value_date

    @cached_property
    def value_date(self) -> date:
        return (self.trade_date - BDay(1)).date()

    @property
    def previous_trade_proposal(self) -> SelfTradeProposal | None:
        future_proposals = TradeProposal.objects.filter(portfolio=self.portfolio).filter(
            trade_date__lt=self.trade_date, status=TradeProposal.Status.APPROVED
        )
        if future_proposals.exists():
            return future_proposals.latest("trade_date")
        return None

    @property
    def next_trade_proposal(self) -> SelfTradeProposal | None:
        future_proposals = TradeProposal.objects.filter(portfolio=self.portfolio).filter(
            trade_date__gt=self.trade_date, status=TradeProposal.Status.APPROVED
        )
        if future_proposals.exists():
            return future_proposals.earliest("trade_date")
        return None

    @property
    def base_assets(self) -> dict[int, Decimal]:
        """
        Return a dictionary representation (instrument_id: target weight) of this trade proposal
        Returns:
            A dictionary representation

        """
        return {
            v["underlying_instrument"]: v["target_weight"]
            for v in self.trades.all()
            .annotate_base_info()
            .filter(status=Trade.Status.EXECUTED)
            .values("underlying_instrument", "target_weight")
        }

    def __str__(self) -> str:
        return f"{self.portfolio.name}: {self.trade_date} ({self.status})"

    def convert_to_portfolio(self, use_effective: bool = False) -> PortfolioDTO:
        """
        Data Transfer Object
        Returns:
            DTO trade object
        """
        portfolio = {}
        for asset in self.portfolio.assets.filter(date=self.last_effective_date):
            portfolio[asset.underlying_quote] = dict(
                shares=asset._shares,
                weighting=asset.weighting,
                delta_weight=Decimal("0"),
                price=asset._price,
                currency_fx_rate=asset._currency_fx_rate,
            )
        for trade in self.trades.all().annotate_base_info():
            portfolio[trade.underlying_instrument] = dict(
                weighting=trade._previous_weight,
                delta_weight=trade.weighting,
                shares=trade._target_shares if not use_effective else trade._effective_shares,
                price=trade.price,
                currency_fx_rate=trade.currency_fx_rate,
            )

        previous_weights = dict(map(lambda r: (r[0].id, float(r[1]["weighting"])), portfolio.items()))
        try:
            drifted_weights = self.portfolio.get_analytic_portfolio(
                self.value_date, weights=previous_weights
            ).get_next_weights()
        except InvalidAnalyticPortfolio:
            drifted_weights = {}
        positions = []
        for instrument, row in portfolio.items():
            weighting = row["weighting"]
            try:
                drift_factor = Decimal(drifted_weights.pop(instrument.id)) / weighting if weighting else Decimal("1")
            except KeyError:
                drift_factor = Decimal("1")
            if not use_effective:
                weighting = weighting * drift_factor + row["delta_weight"]
            positions.append(
                PositionDTO(
                    underlying_instrument=instrument.id,
                    instrument_type=instrument.instrument_type.id,
                    weighting=weighting,
                    drift_factor=drift_factor if use_effective else Decimal("1"),
                    shares=row["shares"],
                    currency=instrument.currency.id,
                    date=self.last_effective_date if use_effective else self.trade_date,
                    is_cash=instrument.is_cash or instrument.is_cash_equivalent,
                    price=row["price"],
                    currency_fx_rate=row["currency_fx_rate"],
                )
            )
        return PortfolioDTO(positions)

    # Start tools methods
    def _clone(self, **kwargs) -> SelfTradeProposal:
        """
        Method to clone self as a new trade proposal. It will automatically shift the trade date if a proposal already exists
        Args:
            **kwargs: The keyword arguments
        Returns:
            The cloned trade proposal
        """
        trade_date = kwargs.get("clone_date", self.trade_date)

        # Find the next valid trade date
        while TradeProposal.objects.filter(portfolio=self.portfolio, trade_date=trade_date).exists():
            trade_date += timedelta(days=1)

        trade_proposal_clone = TradeProposal.objects.create(
            trade_date=trade_date,
            comment=kwargs.get("clone_comment", self.comment),
            status=TradeProposal.Status.DRAFT,
            rebalancing_model=self.rebalancing_model,
            portfolio=self.portfolio,
            creator=self.creator,
        )
        for trade in self.trades.all():
            trade.id = None
            trade.trade_proposal = trade_proposal_clone
            trade.save()

        return trade_proposal_clone

    def normalize_trades(self, total_target_weight: Decimal = Decimal("1.0")):
        """
        Call the trading service with the existing trades and normalize them in order to obtain a total sum target weight of 100%
        The existing trade will be modified directly with the given normalization factor
        """
        service = TradingService(
            self.trade_date,
            effective_portfolio=self._get_default_effective_portfolio(),
            target_portfolio=self.convert_to_portfolio(),
            total_target_weight=total_target_weight,
        )
        leftovers_trades = self.trades.all()
        for underlying_instrument_id, trade_dto in service.trades_batch.trades_map.items():
            with suppress(Trade.DoesNotExist):
                trade = self.trades.get(underlying_instrument_id=underlying_instrument_id)
                trade.weighting = round(trade_dto.delta_weight, Trade.TRADE_WEIGHTING_PRECISION)
                trade.save()
                leftovers_trades = leftovers_trades.exclude(id=trade.id)
        leftovers_trades.delete()
        t_weight = self.trades.all().annotate_base_info().aggregate(models.Sum("target_weight"))[
            "target_weight__sum"
        ] or Decimal("0.0")
        # we handle quantization error due to the decimal max digits. In that case, we take the biggest trade (highest weight) and we remove the quantization error
        if quantize_error := (t_weight - total_target_weight):
            biggest_trade = self.trades.latest("weighting")
            biggest_trade.weighting -= quantize_error
            biggest_trade.save()

    def _get_default_target_portfolio(self, **kwargs) -> PortfolioDTO:
        if self.rebalancing_model:
            params = {}
            if rebalancer := getattr(self.portfolio, "automatic_rebalancer", None):
                params.update(rebalancer.parameters)
            params.update(kwargs)
            return self.rebalancing_model.get_target_portfolio(
                self.portfolio, self.trade_date, self.value_date, **params
            )
        if self.trades.exists():
            return self.convert_to_portfolio()
        # Return the current portfolio by default
        return self.convert_to_portfolio(use_effective=False)

    def _get_default_effective_portfolio(self):
        return self.convert_to_portfolio(use_effective=True)

    def reset_trades(
        self,
        target_portfolio: PortfolioDTO | None = None,
        effective_portfolio: PortfolioRole | None = None,
        validate_trade: bool = True,
        total_target_weight: Decimal = Decimal("1.0"),
    ):
        """
        Will delete all existing trades and recreate them from the method `create_or_update_trades`
        """
        if self.rebalancing_model:
            self.trades.all().delete()
        # delete all existing trades
        # Get effective and target portfolio
        if not target_portfolio:
            target_portfolio = self._get_default_target_portfolio()
        if not effective_portfolio:
            effective_portfolio = self._get_default_effective_portfolio()

        if target_portfolio:
            service = TradingService(
                self.trade_date,
                effective_portfolio=effective_portfolio,
                target_portfolio=target_portfolio,
                total_target_weight=total_target_weight,
            )
            if validate_trade:
                service.is_valid()
                trades = service.validated_trades
            else:
                trades = service.trades_batch.trades_map.values()
            for trade_dto in trades:
                instrument = Instrument.objects.get(id=trade_dto.underlying_instrument)
                if not instrument.is_cash:  # we do not save trade that includes cash component
                    currency_fx_rate = instrument.currency.convert(
                        self.value_date, self.portfolio.currency, exact_lookup=True
                    )
                    # we cannot do a bulk-create because Trade is a multi table inheritance
                    weighting = round(trade_dto.delta_weight, Trade.TRADE_WEIGHTING_PRECISION)
                    drift_factor = trade_dto.drift_factor
                    try:
                        trade = self.trades.get(underlying_instrument=instrument)
                        trade.weighting = weighting
                        trade.currency_fx_rate = currency_fx_rate
                        trade.status = Trade.Status.DRAFT
                        trade.drift_factor = drift_factor
                    except Trade.DoesNotExist:
                        trade = Trade(
                            underlying_instrument=instrument,
                            currency=instrument.currency,
                            value_date=self.value_date,
                            transaction_date=self.trade_date,
                            trade_proposal=self,
                            portfolio=self.portfolio,
                            weighting=weighting,
                            drift_factor=drift_factor,
                            status=Trade.Status.DRAFT,
                            currency_fx_rate=currency_fx_rate,
                        )
                    trade.price = trade.get_price()
                    # if we cannot automatically find a price, we consider the stock is invalid and we sell it
                    if not trade.price:
                        trade.price = Decimal("0.0")
                        trade.weighting = -trade_dto.effective_weight

                    trade.save()

    def approve_workflow(
        self,
        approve_automatically: bool = True,
        silent_exception: bool = False,
        force_reset_trade: bool = False,
        broadcast_changes_at_date: bool = True,
        **reset_trades_kwargs,
    ):
        if self.status == TradeProposal.Status.APPROVED:
            logger.info("Reverting trade proposal ...")
            self.revert()
        if self.status == TradeProposal.Status.DRAFT:
            if (
                self.rebalancing_model or force_reset_trade
            ):  # if there is no position (for any reason) or we the trade proposal has a rebalancer model attached (trades are computed based on an aglo), we reapply this trade proposal
                logger.info("Resetting trades ...")
                try:  # we silent any validation error while setting proposal, because if this happens, we assume the current trade proposal state if valid and we continue to batch compute
                    self.reset_trades(**reset_trades_kwargs)
                except (ValidationError, DatabaseError) as e:
                    self.status = TradeProposal.Status.FAILED
                    if not silent_exception:
                        raise ValidationError(e)
                    return
            logger.info("Submitting trade proposal ...")
            self.submit()
        if self.status == TradeProposal.Status.SUBMIT:
            logger.info("Approving trade proposal ...")
            if approve_automatically and self.portfolio.can_be_rebalanced:
                self.approve(replay=False, broadcast_changes_at_date=broadcast_changes_at_date)

    def replay(self, force_reset_trade: bool = False, broadcast_changes_at_date: bool = True):
        last_trade_proposal = self
        last_trade_proposal_created = False
        while last_trade_proposal and last_trade_proposal.status == TradeProposal.Status.APPROVED:
            if not last_trade_proposal_created:
                logger.info(f"Replaying trade proposal {last_trade_proposal}")
                last_trade_proposal.approve_workflow(
                    silent_exception=True,
                    force_reset_trade=force_reset_trade,
                    broadcast_changes_at_date=broadcast_changes_at_date,
                )
                last_trade_proposal.save()
                if last_trade_proposal.status != TradeProposal.Status.APPROVED:
                    break
            next_trade_proposal = last_trade_proposal.next_trade_proposal
            if next_trade_proposal:
                next_trade_date = next_trade_proposal.trade_date - timedelta(days=1)
            elif next_expected_rebalancing_date := self.portfolio.get_next_rebalancing_date(
                last_trade_proposal.trade_date
            ):
                next_trade_date = (
                    next_expected_rebalancing_date + timedelta(days=7)
                )  # we don't know yet if rebalancing is valid and can be executed on `next_expected_rebalancing_date`, so we add safety window of 7 days
            else:
                next_trade_date = date.today()
            next_trade_date = min(next_trade_date, date.today())
            positions, overriding_trade_proposal = self.portfolio.drift_weights(
                last_trade_proposal.trade_date, next_trade_date, stop_at_rebalancing=True
            )

            # self.portfolio.assets.filter(
            #     date__gt=last_trade_proposal.trade_date, date__lte=next_trade_date, is_estimated=False
            # ).update(
            #     is_estimated=True
            # )  # ensure that we reset non estimated position leftover to estimated between trade proposal during replay
            self.portfolio.bulk_create_positions(
                positions,
                delete_leftovers=True,
                compute_metrics=False,
                broadcast_changes_at_date=broadcast_changes_at_date,
                evaluate_rebalancer=False,
            )
            for draft_tp in TradeProposal.objects.filter(
                portfolio=self.portfolio,
                trade_date__gt=last_trade_proposal.trade_date,
                trade_date__lte=next_trade_date,
                status=TradeProposal.Status.DRAFT,
            ):
                draft_tp.reset_trades()
            if overriding_trade_proposal:
                last_trade_proposal_created = True
                last_trade_proposal = overriding_trade_proposal
            else:
                last_trade_proposal_created = False
                last_trade_proposal = next_trade_proposal

    def invalidate_future_trade_proposal(self):
        # Delete all future automatic trade proposals and set the manual one into a draft state
        self.portfolio.trade_proposals.filter(
            trade_date__gt=self.trade_date, rebalancing_model__isnull=False, comment="Automatic rebalancing"
        ).delete()
        for future_trade_proposal in self.portfolio.trade_proposals.filter(
            trade_date__gt=self.trade_date, status=TradeProposal.Status.APPROVED
        ):
            future_trade_proposal.revert()
            future_trade_proposal.save()

    def get_estimated_shares(
        self, weight: Decimal, underlying_quote: Instrument, quote_price: Decimal
    ) -> Decimal | None:
        """
        Estimates the number of shares for a trade based on the given weight and underlying quote.

        This method calculates the estimated shares by dividing the trade's total value in the portfolio's currency by the price of the underlying quote in the same currency. It handles currency conversion and suppresses any ValueError that might occur during the price retrieval.

        Args:
            weight (Decimal): The weight of the trade.
            underlying_quote (Instrument): The underlying instrument for the trade.

        Returns:
            Decimal | None: The estimated number of shares or None if the calculation fails.
        """
        # Retrieve the price of the underlying quote on the trade date TODO: this is very slow and probably due to the to_date argument to the dl which slowdown drastically the query

        # Calculate the trade's total value in the portfolio's currency
        trade_total_value_fx_portfolio = self.portfolio_total_asset_value * weight

        # Convert the quote price to the portfolio's currency
        price_fx_portfolio = quote_price * underlying_quote.currency.convert(
            self.trade_date, self.portfolio.currency, exact_lookup=False
        )

        # If the price is valid, calculate and return the estimated shares
        if price_fx_portfolio:
            return trade_total_value_fx_portfolio / price_fx_portfolio

    def get_round_lot_size(self, shares: Decimal, underlying_quote: Instrument) -> Decimal:
        if (round_lot_size := underlying_quote.round_lot_size) != 1 and (
            not underlying_quote.exchange or underlying_quote.exchange.apply_round_lot_size
        ):
            if shares > 0:
                shares = math.ceil(shares / round_lot_size) * round_lot_size
            elif abs(shares) > round_lot_size:
                shares = math.floor(shares / round_lot_size) * round_lot_size
        return shares

    def get_estimated_target_cash(self, currency: Currency) -> AssetPosition:
        """
        Estimates the target cash weight and shares for a trade proposal.

        This method calculates the target cash weight by summing the weights of cash trades and adding any leftover weight from non-cash trades. It then estimates the target shares for this cash component if the portfolio is not only weighting-based.

        Args:
            currency (Currency): The currency for the target currency component

        Returns:
            tuple[Decimal, Decimal]: A tuple containing the target cash weight and the estimated target shares.
        """
        # Retrieve trades with base information
        trades = self.trades.all().annotate_base_info()

        # Calculate the target cash weight from cash trades
        target_cash_weight = trades.filter(
            underlying_instrument__is_cash=True, underlying_instrument__currency=currency
        ).aggregate(s=models.Sum("target_weight"))["s"] or Decimal(0)
        # if the specified currency match the portfolio's currency, we include the weight leftover to this cash compoenent
        if currency == self.portfolio.currency:
            # Calculate the total target weight of all trades
            total_target_weight = trades.aggregate(s=models.Sum("target_weight"))["s"] or Decimal(0)

            # Add any leftover weight as cash
            target_cash_weight += Decimal(1) - total_target_weight

        # Initialize target shares to zero
        total_target_shares = Decimal(0)

        # If the portfolio is not only weighting-based, estimate the target shares for the cash component
        if not self.portfolio.only_weighting:
            # Get or create a cash component for the portfolio's currency
            cash_component = Cash.objects.get_or_create(
                currency=currency, defaults={"is_cash": True, "name": currency.title}
            )[0]

            # Estimate the target shares for the cash component
            with suppress(ValueError):
                total_target_shares = self.get_estimated_shares(target_cash_weight, cash_component, Decimal("1.0"))

        cash_component = Cash.objects.get_or_create(
            currency=self.portfolio.currency, defaults={"name": self.portfolio.currency.title}
        )[0]
        # otherwise, we create a new position
        underlying_quote_price = InstrumentPrice.objects.get_or_create(
            instrument=cash_component,
            date=self.trade_date,
            calculated=False,
            defaults={"net_value": Decimal(1)},
        )[0]
        return AssetPosition(
            underlying_quote=cash_component,
            portfolio_created=None,
            portfolio=self.portfolio,
            date=self.trade_date,
            weighting=target_cash_weight,
            initial_price=underlying_quote_price.net_value,
            initial_shares=total_target_shares,
            asset_valuation_date=self.trade_date,
            underlying_quote_price=underlying_quote_price,
            currency=cash_component.currency,
            is_estimated=False,
        )

    # Start FSM logics

    @transition(
        field=status,
        source=Status.DRAFT,
        target=Status.SUBMIT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label="Submit",
                action_label="Submit",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def submit(self, by=None, description=None, **kwargs):
        trades = []
        trades_validation_warnings = []
        for trade in self.trades.all():
            trade_warnings = trade.submit(
                by=by, description=description, portfolio_total_asset_value=self.portfolio_total_asset_value, **kwargs
            )
            if trade_warnings:
                trades_validation_warnings.extend(trade_warnings)
            trades.append(trade)

        Trade.objects.bulk_update(trades, ["status", "shares", "weighting"])

        # If we estimate cash on this trade proposal, we make sure to create the corresponding cash component
        estimated_cash_position = self.get_estimated_target_cash(self.portfolio.currency)
        target_portfolio = self.validated_trading_service.trades_batch.convert_to_portfolio(
            estimated_cash_position._build_dto()
        )
        self.evaluate_active_rules(self.trade_date, target_portfolio, asynchronously=True)
        return trades_validation_warnings

    def can_submit(self):
        errors = dict()
        errors_list = []
        if self.trades.exists() and self.trades.exclude(status=Trade.Status.DRAFT).exists():
            errors_list.append(_("All trades need to be draft before submitting"))
        service = self.validated_trading_service
        try:
            service.is_valid(ignore_error=True)
            # if service.trades_batch.total_abs_delta_weight == 0:
            #     errors_list.append(
            #         "There is no change detected in this trade proposal. Please submit at last one valid trade"
            #     )
            if len(service.validated_trades) == 0:
                errors_list.append(_("There is no valid trade on this proposal"))
            if service.errors:
                errors_list.extend(service.errors)
            if errors_list:
                errors["non_field_errors"] = errors_list
        except ValidationError:
            errors["non_field_errors"] = service.errors
            with suppress(KeyError):
                del self.__dict__["validated_trading_service"]
        return errors

    @property
    def can_be_approved_or_denied(self):
        return not self.has_non_successful_checks and self.portfolio.is_manageable

    @transition(
        field=status,
        source=Status.SUBMIT,
        target=Status.APPROVED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.can_be_approved_or_denied,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.APPROVE.icon,
                key="approve",
                label="Approve",
                action_label="Approve",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def approve(self, by=None, description=None, replay: bool = True, **kwargs):
        # We validate trade which will create or update the initial asset positions
        if not self.portfolio.can_be_rebalanced:
            raise ValueError("Non-Rebalanceable portfolio cannot be traded manually.")
        trades = []
        assets = []
        warnings = []
        # We do not want to create the estimated cash position if there is not trades in the trade proposal (shouldn't be possible anyway)
        estimated_cash_position = self.get_estimated_target_cash(self.portfolio.currency)

        for trade in self.trades.all():
            with suppress(ValueError):
                asset = trade.get_asset()
                # we add the corresponding asset only if it is not the cache position (already included in estimated_cash_position)
                if asset.underlying_quote != estimated_cash_position.underlying_quote:
                    assets.append(asset)
            trade.status = Trade.Status.EXECUTED
            trades.append(trade)

        # if there is cash leftover, we create an extra asset position to hold the cash component
        if estimated_cash_position.weighting and len(trades) > 0:
            warnings.append(
                f"We created automatically a cash position of weight {estimated_cash_position.weighting:.2%}"
            )
            estimated_cash_position.pre_save()
            assets.append(estimated_cash_position)

        Trade.objects.bulk_update(trades, ["status"])
        self.portfolio.bulk_create_positions(
            AssetPositionIterator(self.portfolio).add(assets, is_estimated=False),
            evaluate_rebalancer=False,
            force_save=True,
            **kwargs,
        )
        if replay and self.portfolio.is_manageable:
            replay_as_task.delay(self.id, user_id=by.id if by else None)
        return warnings

    def can_approve(self):
        errors = dict()
        if not self.portfolio.can_be_rebalanced:
            errors["non_field_errors"] = [_("The portfolio does not allow manual rebalanced")]
        if self.trades.exclude(status=Trade.Status.SUBMIT).exists():
            errors["non_field_errors"] = [
                _("At least one trade needs to be submitted to be able to approve this proposal")
            ]
        if not self.portfolio.can_be_rebalanced:
            errors["portfolio"] = [
                [_("The portfolio needs to be a model portfolio in order to approve this trade proposal manually")]
            ]
        if self.has_non_successful_checks:
            errors["non_field_errors"] = [_("The pre trades rules did not passed successfully")]
        return errors

    @transition(
        field=status,
        source=Status.SUBMIT,
        target=Status.DENIED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.can_be_approved_or_denied,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.DENY.icon,
                key="deny",
                label="Deny",
                action_label="Deny",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def deny(self, by=None, description=None, **kwargs):
        self.trades.all().delete()
        with suppress(KeyError):
            del self.__dict__["validated_trading_service"]

    def can_deny(self):
        errors = dict()
        if self.trades.exclude(status=Trade.Status.SUBMIT).exists():
            errors["non_field_errors"] = [
                _("At least one trade needs to be submitted to be able to deny this proposal")
            ]
        return errors

    @transition(
        field=status,
        source=Status.SUBMIT,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.has_all_check_completed
        or not instance.checks.exists(),  # we wait for all checks to succeed before proposing the back to draft transition
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.UNDO.icon,
                key="backtodraft",
                label="Back to Draft",
                action_label="backtodraft",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def backtodraft(self, **kwargs):
        with suppress(KeyError):
            del self.__dict__["validated_trading_service"]
        self.trades.update(status=Trade.Status.DRAFT)
        self.checks.delete()

    def can_backtodraft(self):
        pass

    @transition(
        field=status,
        source=Status.APPROVED,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.REGENERATE.icon,
                key="revert",
                label="Revert",
                action_label="revert",
                description_fields="<p>Unapply trades and move everything back to draft (i.e. The underlying asset positions will change like the trades were never applied)</p>",
            )
        },
    )
    def revert(self, **kwargs):
        with suppress(KeyError):
            del self.__dict__["validated_trading_service"]
        trades = []
        self.portfolio.assets.filter(date=self.trade_date, is_estimated=False).update(
            is_estimated=True
        )  # we delete the existing portfolio as it has been reverted
        for trade in self.trades.all():
            trade.status = Trade.Status.DRAFT
            trade.drift_factor = Decimal("1")
            trades.append(trade)
        Trade.objects.bulk_update(trades, ["status", "drift_factor"])

    def can_revert(self):
        errors = dict()
        if not self.portfolio.can_be_rebalanced:
            errors["portfolio"] = [
                _("The portfolio needs to be a model portfolio in order to revert this trade proposal manually")
            ]
        return errors

    # End FSM logics

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbportfolio:tradeproposal"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbportfolio:tradeproposalrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{_portfolio.name}} ({{trade_date}})"


@receiver(post_save, sender="wbportfolio.TradeProposal")
def post_fail_trade_proposal(sender, instance: TradeProposal, created, raw, **kwargs):
    # if we have a trade proposal in a fail state, we ensure that all future existing trade proposal are either deleted (automatic one) or set back to draft
    if not raw and instance.status == TradeProposal.Status.FAILED:
        # we delete all trade proposal that have a rebalancing model and are marked as "automatic" (quite hardcoded yet)
        instance.invalidate_future_trade_proposal()
        instance.invalidate_future_trade_proposal()


@shared_task(queue="portfolio")
def replay_as_task(trade_proposal_id, user_id: int | None = None):
    trade_proposal = TradeProposal.objects.get(id=trade_proposal_id)
    trade_proposal.replay()
    if user_id:
        user = User.objects.get(id=user_id)
        send_notification(
            code="wbportfolio.portfolio.replay_done",
            title="Trade Proposal Replay Completed",
            body=f'We’ve successfully replayed your trade proposal for "{trade_proposal.portfolio}" from {trade_proposal.trade_date:%Y-%m-%d}. You can now review its updated composition.',
            user=user,
            reverse_name="wbportfolio:portfolio-detail",
            reverse_args=[trade_proposal.portfolio.id],
        )



from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.preview import PreviewViewConfig


class PortfolioPreviewConfig(PreviewViewConfig):
    def get_display(self) -> Display:
        return create_simple_display(
            [
                ["name", "currency", "updated_at"],
                [repeat_field(3, "depends_on")],
            ]
        )



import numpy as np
import pandas as pd

FIELD_MAP = {"Date": "date", "Quantity": "outstanding_shares", "NAV": "net_value"}


def parse(import_source):
    base_filename = import_source.source.import_parameters.get("valuation_base_filename", "StrategicCertificate")
    data = []
    if base_filename in import_source.file.name:
        df = pd.read_excel(import_source.file, engine="openpyxl", sheet_name=0)
        xx, yy = np.where(df == "Date")
        product_ticker = list(filter(lambda x: "Unnamed" not in x, df.columns))
        df = df.iloc[xx[0] :, yy[0] :]

        df = df.rename(columns=df.iloc[0]).drop(df.index[0]).dropna(how="all")
        df = df.rename(columns=FIELD_MAP)
        df = df.drop(columns=df.columns.difference(FIELD_MAP.values()))
        df["date"] = pd.to_datetime(df["date"]).dt.strftime("%Y-%m-%d")
        df["outstanding_shares"] = -df["outstanding_shares"]

        if len(product_ticker) == 0:
            for k, v in import_source.source.import_parameter.items():
                df[k] = v
        else:
            df["instrument__ticker"] = product_ticker[0]
        df = df.replace([np.inf, -np.inf, np.nan], None)
        data = df.to_dict("records")
    return {"data": data}



from contextlib import suppress
from datetime import date
from typing import Generator

import pandas as pd
from django.db.models import Q
from django.utils.functional import cached_property
from wbfdm.contrib.metric.backends.base import Metric
from wbfdm.contrib.metric.decorators import register
from wbfdm.enums import Financial, MarketData
from wbfdm.models import Instrument

from wbportfolio.models import Portfolio

from .base import PortfolioMetricBaseBackend
from .constants import (
    PORTFOLIO_CAPITAL_EMPLOYED,
    PORTFOLIO_EBIT,
    PORTFOLIO_LIABILITIES,
    PORTFOLIO_ROCE,
    PORTFOLIO_TOTAL_ASSETS,
)


class DataLoader:
    def __init__(
        self, portfolio: Portfolio, val_date: date, past_offset: int = 4, forecast_offset: int = 4, freq: str = "YE"
    ):
        self.portfolio = portfolio
        self.val_date = val_date
        self.instruments = Instrument.objects.filter(id__in=self.base_df.index)
        self.past_offset = past_offset
        self.forecast_offset = forecast_offset
        self.pivot_dates = list(
            pd.date_range(end=self.val_date, periods=self.past_offset, freq=freq, inclusive="left")
        ) + list(pd.date_range(start=self.val_date, periods=self.past_offset, freq=freq, inclusive="right"))
        self.columns_map = {
            self.val_date.year + offset: f"fy{offset + 1}".replace("-", "_")
            for offset in range(-1 * self.past_offset, self.forecast_offset)
        }
        self.from_year = self.val_date.year - self.past_offset
        self.to_year = self.val_date.year + self.forecast_offset - 1

        self.empty_series = pd.Series(dtype="float64")

    def _get_financial_df(self, financial: Financial, target_currency: str = "USD", **kwargs) -> pd.DataFrame:
        """
        Private helper method to gather necessary financial data from the finanical dataloader and pivot it into Instrument*Years format

        Args:
            financial: Financial Metric
            target_currency: Target currency
            **kwargs: fdm datalaoder keyword argument

        Returns:
            A dataframe whose index is the instrument ids and columns are the years offsets
        """

        df = pd.DataFrame(
            self.instruments.dl.financials(
                values=[financial],
                from_year=self.from_year,
                to_year=self.to_year,
                target_currency=target_currency,
                **kwargs,
            )
        )
        if df.empty:
            raise ValueError(f"No financial data for {financial.value}")
        return df.pivot_table(index="instrument_id", values="value", columns="year", aggfunc="first")

    # Data Input
    @cached_property
    def market_capitalization_df(self) -> pd.DataFrame:
        df_list = []
        for pivot_date in self.pivot_dates:
            try:
                mkt_caps = pd.DataFrame(
                    self.instruments.dl.market_data(
                        values=[MarketData.MARKET_CAPITALIZATION],
                        from_date=(pivot_date - pd.tseries.offsets.BDay(3)).date(),
                        to_date=pivot_date,
                        target_currency="USD",
                    )
                )
                res = (
                    mkt_caps[["instrument_id", "market_capitalization"]]
                    .groupby("instrument_id")
                    .last()["market_capitalization"]
                ).astype(float)
            except KeyError:
                res = self.empty_series
            res = res.rename(pivot_date.year)
            df_list.append(res)
        return pd.concat(df_list, axis=1).ffill(axis=1)

    @cached_property
    def ebit_usd_df(self) -> pd.DataFrame:
        return self._get_financial_df(Financial.EBIT)

    @cached_property
    def total_assets_usd_df(self) -> pd.DataFrame:
        return self._get_financial_df(Financial.TOTAL_ASSETS)

    @cached_property
    def liabilities_usd_df(self) -> pd.DataFrame:
        return self._get_financial_df(Financial.CURRENT_LIABILITIES)

    # Generate basic portfolio data as a dataframe
    @cached_property
    def base_df(self) -> pd.DataFrame:
        df = pd.DataFrame(
            self.portfolio.assets.filter(date=self.val_date)
            .exclude(Q(underlying_instrument__is_cash=True) | Q(underlying_instrument__is_cash_equivalent=True))
            .values("underlying_instrument", "shares", "price_fx_usd", "total_value_fx_usd", "weighting"),
            columns=["underlying_instrument", "shares", "price_fx_usd", "total_value_fx_usd", "weighting"],
        )
        df = (
            df.groupby("underlying_instrument")
            .agg(
                {
                    "shares": "sum",
                    "price_fx_usd": "first",
                    "total_value_fx_usd": "sum",
                    "weighting": "sum",
                }
            )
            .astype(float)
        )
        df["weighting"] = df["weighting"] / df["weighting"].sum()
        df.index = df.index.rename("instrument_id")
        return df

    # Portfolio Metrics

    @cached_property
    def portfolio_ownership_df(self) -> pd.DataFrame:
        return (1.0 / self.market_capitalization_df).multiply(self.base_df.total_value_fx_usd, axis=0)

    @cached_property
    def portfolio_ebit_usd_df(self) -> pd.DataFrame:
        df = self.portfolio_ownership_df * self.ebit_usd_df
        df.aggregate = df.sum(axis=0)
        return df

    @cached_property
    def portfolio_total_assets_usd_df(self) -> pd.DataFrame:
        df = self.portfolio_ownership_df * self.total_assets_usd_df
        df.aggregate = df.sum(axis=0)
        return df

    @cached_property
    def portfolio_liabilities_usd_df(self) -> pd.DataFrame:
        df = self.portfolio_ownership_df * self.liabilities_usd_df
        df.aggregate = df.sum(axis=0)
        return df

    @cached_property
    def portfolio_aggregation_capital_employed(self) -> pd.Series:
        return self.portfolio_total_assets_usd_df.aggregate - self.portfolio_liabilities_usd_df.aggregate

    @cached_property
    def portfolio_aggregation_roce(self) -> pd.Series:
        return self.portfolio_ebit_usd_df.aggregate / self.portfolio_aggregation_capital_employed.rolling(2).mean()


@register()
class PortfolioBaseMetricBackend(PortfolioMetricBaseBackend):
    portfolio_ebit = PORTFOLIO_EBIT
    portfolio_total_assets = PORTFOLIO_TOTAL_ASSETS
    portfolio_liabilities = PORTFOLIO_LIABILITIES
    portfolio_capital_employed = PORTFOLIO_CAPITAL_EMPLOYED
    portfolio_roce = PORTFOLIO_ROCE

    keys = [
        PORTFOLIO_EBIT,
        PORTFOLIO_TOTAL_ASSETS,
        PORTFOLIO_LIABILITIES,
        PORTFOLIO_CAPITAL_EMPLOYED,
        PORTFOLIO_ROCE,
    ]

    def compute_metrics(self, basket: Portfolio) -> Generator[Metric, None, None]:
        val_date = self._get_valid_date(basket)
        dataloader = DataLoader(basket, val_date)

        # Load all necessary metrics into a Metric DTO from the dataloader
        with suppress(ValueError):
            portfolio_ebit_metrics = self.convert_df_into_metrics(
                dataloader.portfolio_ebit_usd_df,
                "portfolio_ebit",
                dataloader.portfolio.id,
                val_date,
                columns_map=dataloader.columns_map,
            )
            portfolio_aggregation_ebit_metrics = self.convert_df_into_metrics(
                dataloader.portfolio_ebit_usd_df.aggregate,
                "portfolio_ebit",
                dataloader.portfolio.id,
                val_date,
                columns_map=dataloader.columns_map,
                dependency_metrics=portfolio_ebit_metrics,
            )

            portfolio_total_assets_metrics = self.convert_df_into_metrics(
                dataloader.portfolio_total_assets_usd_df,
                "portfolio_total_assets",
                dataloader.portfolio.id,
                val_date,
                columns_map=dataloader.columns_map,
            )
            portfolio_aggregation_total_assets_metrics = self.convert_df_into_metrics(
                dataloader.portfolio_total_assets_usd_df.aggregate,
                "portfolio_total_assets",
                dataloader.portfolio.id,
                val_date,
                columns_map=dataloader.columns_map,
                dependency_metrics=portfolio_total_assets_metrics,
            )

            portfolio_liabilities_metrics = self.convert_df_into_metrics(
                dataloader.portfolio_liabilities_usd_df,
                "portfolio_liabilities",
                dataloader.portfolio.id,
                val_date,
                columns_map=dataloader.columns_map,
            )
            portfolio_aggregation_liabilities_metrics = self.convert_df_into_metrics(
                dataloader.portfolio_liabilities_usd_df.aggregate,
                "portfolio_liabilities",
                dataloader.portfolio.id,
                val_date,
                columns_map=dataloader.columns_map,
                dependency_metrics=portfolio_liabilities_metrics,
            )

            portfolio_aggregation_capital_metrics = self.convert_df_into_metrics(
                dataloader.portfolio_aggregation_capital_employed,
                "portfolio_capital_employed",
                dataloader.portfolio.id,
                val_date,
                columns_map=dataloader.columns_map,
                dependency_metrics=portfolio_aggregation_total_assets_metrics
                + portfolio_aggregation_liabilities_metrics,
            )

            portfolio_aggregation_roce_metrics = self.convert_df_into_metrics(
                dataloader.portfolio_aggregation_roce,
                "portfolio_roce",
                dataloader.portfolio.id,
                val_date,
                columns_map=dataloader.columns_map,
                dependency_metrics=portfolio_aggregation_ebit_metrics + portfolio_aggregation_capital_metrics,
            )

            yield from portfolio_aggregation_roce_metrics



from decimal import Decimal

import pandas as pd
import plotly.graph_objects as go
from django.contrib.messages import error, warning
from django.db.models import (
    BooleanField,
    Case,
    DecimalField,
    ExpressionWrapper,
    F,
    OuterRef,
    Subquery,
    Sum,
    Value,
    When,
)
from django.db.models.functions import Coalesce
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from django_filters.rest_framework import DjangoFilterBackend
from wbcore import viewsets
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbcore.utils.strings import format_number
from wbcore.viewsets.mixins import OrderableMixin
from wbcrm.models import Account

from wbportfolio.filters import (
    SubscriptionRedemptionFilterSet,
    SubscriptionRedemptionPortfolioFilterSet,
    TradeFilter,
    TradeInstrumentFilterSet,
    TradePortfolioFilter,
)
from wbportfolio.import_export.resources.trades import TradeProposalTradeResource
from wbportfolio.models import Trade, TradeProposal
from wbportfolio.models.transactions.claim import Claim
from wbportfolio.serializers import (
    ReadOnlyTradeTradeProposalModelSerializer,
    TradeModelSerializer,
    TradeProposalRepresentationSerializer,
    TradeRepresentationSerializer,
    TradeTradeProposalModelSerializer,
)
from wbportfolio.viewsets.configs.titles.trades import (
    CustomerDistributionInstrumentTitleConfig,
)

from ..configs import (
    CustodianDistributionInstrumentEndpointConfig,
    CustodianDistributionInstrumentTitleConfig,
    SubscriptionRedemptionDisplayConfig,
    SubscriptionRedemptionEndpointConfig,
    SubscriptionRedemptionTitleConfig,
    TradeButtonConfig,
    TradeDisplayConfig,
    TradeEndpointConfig,
    TradeInstrumentButtonConfig,
    TradeInstrumentEndpointConfig,
    TradeInstrumentTitleConfig,
    TradePortfolioDisplayConfig,
    TradePortfolioEndpointConfig,
    TradePortfolioTitleConfig,
    TradeTitleConfig,
    TradeTradeProposalButtonConfig,
    TradeTradeProposalDisplayConfig,
    TradeTradeProposalEndpointConfig,
)
from ..mixins import UserPortfolioRequestPermissionMixin


class TradeProposalRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:trade"
    queryset = TradeProposal.objects.all()
    serializer_class = TradeProposalRepresentationSerializer


class TradeRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    filterset_class = TradeFilter

    ordering_fields = ("transaction_date", "shares")
    ordering = ["-transaction_date"]
    search_fields = (
        "underlying_instrument__name",
        "bank",
        "shares",
        "comment",
        "external_id",
        "register__register_name_1",
        "register__register_name_2",
        "register__custodian_name_1",
        "register__custodian_name_2",
    )

    queryset = Trade.objects.select_related("underlying_instrument")
    serializer_class = TradeRepresentationSerializer

    def get_queryset(self):
        return (
            Trade.objects.filter(marked_for_deletion=False, pending=False)
            .select_related("portfolio")
            .select_related("underlying_instrument")
            .prefetch_related("claims")
        )


class TradeModelViewSet(UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, viewsets.ModelViewSet):
    IDENTIFIER = "wbportfolio:trade"

    ordering_fields = (
        "id",
        "transaction_subtype",
        "transaction_date",
        "underlying_instrument__name",
        "shares",
        "approved_claimed_shares",
        "pending_claimed_shares",
        "price",
        "total_value",
        "total_value_fx_portfolio",
        "total_value_usd",
        "currency_fx_rate",
        "currency__key",
        "bank",
        "register__computed_str",
        "comment",
    )
    search_fields = (
        "portfolio__name",
        "underlying_instrument__name",
        "bank",
        "comment",
        "external_id",
        "register__computed_str",
        "register__register_name_1",
        "register__register_name_2",
        "register__custodian_name_1",
        "register__custodian_name_2",
    )

    filterset_class = TradeFilter
    queryset = Trade.objects.all()
    serializer_class = TradeModelSerializer

    display_config_class = TradeDisplayConfig
    endpoint_config_class = TradeEndpointConfig
    title_config_class = TradeTitleConfig
    button_config_class = TradeButtonConfig

    def get_aggregates(self, queryset, paginated_queryset):
        return {
            "shares": {
                "Σ": format_number(queryset.aggregate(s=Sum("shares"))["s"] or Decimal(0)),
            },
            "total_value_usd": {
                "Σ": format_number(queryset.aggregate(s=Sum("total_value_usd"))["s"] or Decimal(0)),
            },
        }

    def get_queryset(self):
        qs = super().get_queryset()
        if not self.is_portfolio_manager:
            qs = qs.filter(transaction_subtype__in=[Trade.Type.REDEMPTION, Trade.Type.SUBSCRIPTION])
        qs = (
            qs.annotate(
                fx_rate=CurrencyFXRates.get_fx_rates_subquery(
                    "transaction_date", currency="currency", lookup_expr="exact"
                ),  # this slow down the request. An alternative would be to store the value in the model.
                total_value_usd=ExpressionWrapper(F("total_value"), output_field=DecimalField()),
                total_value_gross_usd=ExpressionWrapper(F("total_value_gross"), output_field=DecimalField()),
                approved_claimed_shares=Coalesce(
                    Subquery(
                        Claim.objects.filter(status=Claim.Status.APPROVED, trade=OuterRef("pk"))
                        .values("trade")
                        .annotate(sum_shares=Sum("shares"))
                        .values("sum_shares")[:1]
                    ),
                    Decimal(0.0),
                ),
                pending_claimed_shares=Coalesce(
                    Subquery(
                        Claim.objects.filter(
                            status__in=[Claim.Status.PENDING, Claim.Status.DRAFT], trade=OuterRef("pk")
                        )
                        .values("trade")
                        .annotate(sum_shares=Sum("shares"))
                        .values("sum_shares")[:1]
                    ),
                    Decimal(0.0),
                ),
                completely_claimed=Case(
                    When(shares=F("approved_claimed_shares"), then=Value(True)),
                    default=Value(False),
                    output_field=BooleanField(),
                ),
                completely_claimed_if_approved=Case(
                    When(completely_claimed=True, then=Value(False)),
                    When(shares=F("approved_claimed_shares") + F("pending_claimed_shares"), then=Value(True)),
                    default=Value(False),
                    output_field=BooleanField(),
                ),
            )
            .select_related("underlying_instrument")
            .select_related("register")
            .select_related("portfolio")
            .select_related("currency")
            .select_related("custodian")
            .select_related("import_source")
            .prefetch_related("claims")
            .prefetch_related("claims__account")
            .prefetch_related("claims__claimant")
            .prefetch_related("claims__product")
        )
        return qs


class SubscriptionRedemptionModelViewSet(TradeModelViewSet):
    IDENTIFIER = "wbportoflio:subscriptionredemption"
    filterset_class = SubscriptionRedemptionFilterSet
    display_config_class = SubscriptionRedemptionDisplayConfig
    title_config_class = SubscriptionRedemptionTitleConfig
    endpoint_config_class = SubscriptionRedemptionEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(transaction_subtype__in=[Trade.Type.REDEMPTION, Trade.Type.SUBSCRIPTION])


class TradePortfolioModelViewSet(TradeModelViewSet):
    IDENTIFIER = "wbportfolio:trade"
    filterset_class = TradePortfolioFilter
    title_config_class = TradePortfolioTitleConfig
    endpoint_config_class = TradePortfolioEndpointConfig
    display_config_class = TradePortfolioDisplayConfig

    def get_aggregates(self, queryset, paginated_queryset):
        if queryset.exists():
            return {
                "total_value_fx_portfolio": {
                    "Σ": format_number(queryset.aggregate(s=Sum(F("total_value_fx_portfolio")))["s"])
                },
                **super().get_aggregates(queryset, paginated_queryset),
            }
        return dict()

    def get_queryset(self):
        return super().get_queryset().filter(portfolio=self.portfolio)


class TradeInstrumentModelViewSet(TradeModelViewSet):
    IDENTIFIER = "wbportfolio:trade"

    filterset_class = TradeInstrumentFilterSet

    title_config_class = TradeInstrumentTitleConfig
    endpoint_config_class = TradeInstrumentEndpointConfig
    button_config_class = TradeInstrumentButtonConfig

    def get_aggregates(self, queryset, paginated_queryset):
        if queryset.exists():
            return {
                "total_value": {"Σ": format_number(queryset.aggregate(s=Sum(F("total_value")))["s"])},
                **super().get_aggregates(queryset, paginated_queryset),
            }
        return dict()

    def get_queryset(self):
        return (
            super().get_queryset().filter(underlying_instrument__in=self.instrument.get_descendants(include_self=True))
        )


class SubscriptionRedemptionInstrumentModelViewSet(SubscriptionRedemptionModelViewSet):
    filterset_class = SubscriptionRedemptionPortfolioFilterSet

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(
                underlying_instrument__in=self.instrument.get_descendants(include_self=True),
                transaction_subtype__in=[Trade.Type.REDEMPTION, Trade.Type.SUBSCRIPTION],
            )
        )


class CustodianDistributionInstrumentChartViewSet(UserPortfolioRequestPermissionMixin, viewsets.ChartViewSet):
    IDENTIFIER = "wbportfolio:custodiandistribution"
    filterset_fields = {
        "transaction_date": ["lte"],
    }
    filter_backends = (DjangoFilterBackend,)
    queryset = Trade.objects.all()

    title_config_class = CustodianDistributionInstrumentTitleConfig
    endpoint_config_class = CustodianDistributionInstrumentEndpointConfig

    def get_plotly(self, queryset):
        fig = go.Figure()
        if queryset.exists():
            df = pd.DataFrame(queryset.values("custodian__name", "custodian__id", "shares"))
            df = df.groupby("custodian__id").agg({"custodian__name": "first", "shares": "sum"})
            fig = go.Figure(data=[go.Pie(labels=df.custodian__name, values=df.shares)])
        return fig

    def get_queryset(self):
        return Trade.objects.filter(underlying_instrument=self.instrument)


class CustomerDistributionInstrumentChartViewSet(UserPortfolioRequestPermissionMixin, viewsets.ChartViewSet):
    IDENTIFIER = "wbportfolio:custodiandistribution"
    filterset_fields = {
        "transaction_date": ["lte"],
    }
    filter_backends = (DjangoFilterBackend,)
    queryset = Trade.objects.all()

    title_config_class = CustomerDistributionInstrumentTitleConfig
    endpoint_config_class = CustodianDistributionInstrumentEndpointConfig

    # TODO: Consider moving into a helper class?
    def group_smaller_items(self, df, threshold=0.01, text="Others"):
        df["weight"] = df.sum_shares / df.sum_shares.sum()
        group_sum = df.loc[df["weight"] < threshold].sum().to_frame().T
        group_sum["account_name"] = text
        rest = df.loc[df["weight"] >= threshold]
        return pd.concat([group_sum, rest])

    def get_plotly(self, queryset):
        fig = go.Figure()
        if queryset.exists():
            df = self.group_smaller_items(
                pd.DataFrame(queryset.values("account_name", "sum_shares")),
            )

            fig = go.Figure(data=[go.Pie(labels=df.account_name, values=df.sum_shares)])
        return fig

    def get_queryset(self):
        return (
            Trade.objects.filter(
                underlying_instrument=self.instrument, transaction_subtype__in=["SUBSCRIPTION", "REDEMPTION"]
            )
            .values("claims")
            .annotate(
                shares=F("claims__shares"),
                account_tree_id=F("claims__account__tree_id"),
                status=F("claims__status"),
            )
            .filter(status="APPROVED")
            .values("account_tree_id")
            .annotate(
                account_name=Subquery(
                    Account.objects.filter(tree_id=OuterRef("account_tree_id"), level=0).values("title")[:1]
                ),
                sum_shares=Sum("shares"),
            )
            .filter(account_name__isnull=False)
        )


class TradeTradeProposalModelViewSet(
    UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, OrderableMixin, viewsets.ModelViewSet
):
    IMPORT_ALLOWED = True
    ordering = (
        "trade_proposal",
        "order",
    )
    ordering_fields = (
        "underlying_instrument__name",
        "underlying_instrument_isin",
        "underlying_instrument_ticker",
        "underlying_instrument_refinitiv_identifier_code",
        "underlying_instrument_instrument_type",
        "target_weight",
        "effective_weight",
        "effective_shares",
        "target_shares",
        "shares",
        "weighting",
    )
    IDENTIFIER = "wbportfolio:tradeproposal"
    search_fields = ("underlying_instrument__name",)
    filterset_fields = {"status": ["exact"]}
    queryset = Trade.objects.all()

    display_config_class = TradeTradeProposalDisplayConfig
    endpoint_config_class = TradeTradeProposalEndpointConfig
    serializer_class = TradeTradeProposalModelSerializer
    button_config_class = TradeTradeProposalButtonConfig

    @cached_property
    def trade_proposal(self):
        return get_object_or_404(TradeProposal, pk=self.kwargs["trade_proposal_id"])

    @cached_property
    def portfolio_total_asset_value(self):
        return self.trade_proposal.portfolio_total_asset_value

    def has_import_permission(self, request) -> bool:  # allow import only on draft trade proposal
        return super().has_import_permission(request) and self.trade_proposal.status == TradeProposal.Status.DRAFT

    def get_import_resource_kwargs(self):
        resource_kwargs = super().get_import_resource_kwargs()
        resource_kwargs["columns_mapping"] = {"underlying_instrument": "underlying_instrument__isin"}
        return resource_kwargs

    def get_resource_class(self):
        return TradeProposalTradeResource

    def get_aggregates(self, queryset, *args, **kwargs):
        agg = {}
        if queryset.exists():
            noncash_aggregates = queryset.filter(underlying_instrument__is_cash=False).aggregate(
                sum_target_weight=Sum(F("target_weight")),
                sum_effective_weight=Sum(F("effective_weight")),
                sum_target_total_value_fx_portfolio=Sum(F("target_total_value_fx_portfolio")),
                sum_effective_total_value_fx_portfolio=Sum(F("effective_total_value_fx_portfolio")),
            )

            # weights aggregates
            cash_sum_effective_weight = Decimal("1.0") - noncash_aggregates["sum_effective_weight"]
            cash_sum_target_cash_weight = Decimal("1.0") - noncash_aggregates["sum_target_weight"]
            noncash_sum_effective_weight = noncash_aggregates["sum_effective_weight"] or Decimal(0)
            noncash_sum_target_weight = noncash_aggregates["sum_target_weight"] or Decimal(0)
            sum_buy_weight = queryset.filter(weighting__gte=0).aggregate(s=Sum(F("weighting")))["s"] or Decimal(0)
            sum_sell_weight = queryset.filter(weighting__lt=0).aggregate(s=Sum(F("weighting")))["s"] or Decimal(0)

            # shares aggregates
            cash_sum_effective_total_value_fx_portfolio = cash_sum_effective_weight * self.portfolio_total_asset_value
            cash_sum_target_total_value_fx_portfolio = cash_sum_target_cash_weight * self.portfolio_total_asset_value
            noncash_sum_effective_total_value_fx_portfolio = noncash_aggregates[
                "sum_effective_total_value_fx_portfolio"
            ] or Decimal(0)
            noncash_sum_target_total_value_fx_portfolio = noncash_aggregates[
                "sum_target_total_value_fx_portfolio"
            ] or Decimal(0)
            sum_buy_total_value_fx_portfolio = queryset.filter(total_value_fx_portfolio__gte=0).aggregate(
                s=Sum(F("total_value_fx_portfolio"))
            )["s"] or Decimal(0)
            sum_sell_total_value_fx_portfolio = queryset.filter(total_value_fx_portfolio__lt=0).aggregate(
                s=Sum(F("total_value_fx_portfolio"))
            )["s"] or Decimal(0)

            agg = {
                "effective_weight": {
                    "Cash": format_number(cash_sum_effective_weight, decimal=6),
                    "Non-Cash": format_number(noncash_sum_effective_weight, decimal=6),
                    "Total": format_number(noncash_sum_effective_weight + cash_sum_effective_weight, decimal=6),
                },
                "target_weight": {
                    "Cash": format_number(cash_sum_target_cash_weight, decimal=6),
                    "Non-Cash": format_number(noncash_sum_target_weight, decimal=6),
                    "Total": format_number(cash_sum_target_cash_weight + noncash_sum_target_weight, decimal=6),
                },
                "effective_total_value_fx_portfolio": {
                    "Cash": format_number(cash_sum_effective_total_value_fx_portfolio, decimal=6),
                    "Non-Cash": format_number(noncash_sum_effective_total_value_fx_portfolio, decimal=6),
                    "Total": format_number(
                        cash_sum_effective_total_value_fx_portfolio + noncash_sum_effective_total_value_fx_portfolio,
                        decimal=6,
                    ),
                },
                "target_total_value_fx_portfolio": {
                    "Cash": format_number(cash_sum_target_total_value_fx_portfolio, decimal=6),
                    "Non-Cash": format_number(noncash_sum_target_total_value_fx_portfolio, decimal=6),
                    "Total": format_number(
                        cash_sum_target_total_value_fx_portfolio + noncash_sum_target_total_value_fx_portfolio,
                        decimal=6,
                    ),
                },
                "weighting": {
                    "Cash Flow": format_number(cash_sum_target_cash_weight - cash_sum_effective_weight, decimal=6),
                    "Buy": format_number(sum_buy_weight, decimal=6),
                    "Sell": format_number(sum_sell_weight, decimal=6),
                },
                "total_value_fx_portfolio": {
                    "Cash Flow": format_number(
                        cash_sum_target_total_value_fx_portfolio - cash_sum_effective_total_value_fx_portfolio,
                        decimal=6,
                    ),
                    "Buy": format_number(sum_buy_total_value_fx_portfolio, decimal=6),
                    "Sell": format_number(sum_sell_total_value_fx_portfolio, decimal=6),
                },
            }

        return agg

    def get_serializer_class(self):
        if self.trade_proposal.status != TradeProposal.Status.DRAFT:
            return ReadOnlyTradeTradeProposalModelSerializer
        return TradeTradeProposalModelSerializer

    def add_messages(self, request, queryset=None, paginated_queryset=None, instance=None, initial=False):
        if queryset is not None and queryset.exists() and self.trade_proposal.status != TradeProposal.Status.APPROVED:
            total_target_weight = queryset.aggregate(c=Sum(F("target_weight")))["c"] or Decimal(0)
            if round(total_target_weight, 3) != 1:
                warning(
                    request,
                    "The total target weight does not equal 1. To avoid automatic cash allocation, please adjust the trade weights to sum up to 1. Otherwise, a cash component will be added when this trade proposal is submitted.",
                )
            if queryset.filter(status=Trade.Status.FAILED):
                error(
                    request,
                    "Some trades failed preparation. To resolve this, please revert the trade proposal to draft, review and correct the trades, and then resubmit.",
                )

    def get_queryset(self):
        if self.is_portfolio_manager:
            qs = super().get_queryset().filter(trade_proposal=self.kwargs["trade_proposal_id"]).annotate_base_info()
        else:
            qs = TradeProposal.objects.none()

        return qs.annotate(
            underlying_instrument_isin=F("underlying_instrument__isin"),
            underlying_instrument_ticker=F("underlying_instrument__ticker"),
            underlying_instrument_refinitiv_identifier_code=F("underlying_instrument__refinitiv_identifier_code"),
            underlying_instrument_instrument_type=Case(
                When(
                    underlying_instrument__parent__is_security=True,
                    then=F("underlying_instrument__parent__instrument_type__short_name"),
                ),
                default=F("underlying_instrument__instrument_type__short_name"),
            ),
            effective_total_value_fx_portfolio=F("effective_weight") * Value(self.portfolio_total_asset_value),
            target_total_value_fx_portfolio=F("target_weight") * Value(self.portfolio_total_asset_value),
            portfolio_currency=F("portfolio__currency__symbol"),
            security=F("underlying_instrument__parent"),
            company=F("underlying_instrument__parent__parent"),
        )



from wbcore import filters as wb_filters
from wbcore.filters.defaults import current_quarter_date_range
from wbcore.pandas.filterset import PandasFilterSetMixin

from wbportfolio.models import Fees

from .utils import get_transaction_default_date_range


class FeesFilter(wb_filters.FilterSet):
    """FilterSet for Fees

    Currently filters:
    - date: daterange

    """

    fee_date = wb_filters.DateRangeFilter(
        method=wb_filters.DateRangeFilter.base_date_range_filter_method,
        label="Date Range",
        initial=get_transaction_default_date_range,
        required=True,
    )

    class Meta:
        model = Fees
        fields = {
            "transaction_subtype": ["exact"],
            "currency_fx_rate": ["gte", "exact", "lte"],
            "product": ["exact"],
            "currency": ["exact"],
            "total_value": ["gte", "exact", "lte"],
            "total_value_fx_portfolio": ["gte", "exact", "lte"],
        }


class FeesProductFilterSet(FeesFilter):
    portfolio = None

    class Meta:
        model = Fees
        fields = {
            "calculated": ["exact"],
            "transaction_subtype": ["exact"],
            "currency_fx_rate": ["gte", "exact", "lte"],
            "currency": ["exact"],
            "total_value": ["gte", "exact", "lte"],
            "total_value_fx_portfolio": ["gte", "exact", "lte"],
        }


class FeesAggregatedFilter(PandasFilterSetMixin, wb_filters.FilterSet):
    fee_date = wb_filters.DateRangeFilter(
        label="Date Range",
        method=wb_filters.DateRangeFilter.base_date_range_filter_method,
        required=True,
        clearable=False,
        initial=current_quarter_date_range,
    )

    class Meta:
        model = Fees
        fields = {}



from decimal import Decimal

from wbfdm.models import InstrumentPrice

from wbportfolio.pms.typing import Portfolio
from wbportfolio.rebalancing.base import AbstractRebalancingModel
from wbportfolio.rebalancing.decorators import register


@register("Equally Weighted Rebalancing")
class EquallyWeightedRebalancing(AbstractRebalancingModel):
    def __init__(self, *args, **kwargs):
        super(EquallyWeightedRebalancing, self).__init__(*args, **kwargs)
        if not self.effective_portfolio:
            self.effective_portfolio = self.portfolio._build_dto(self.trade_date)

    def is_valid(self) -> bool:
        return (
            len(self.effective_portfolio.positions) > 0
            and InstrumentPrice.objects.filter(
                date=self.trade_date, instrument__in=self.effective_portfolio.positions_map.keys()
            ).exists()
        )

    def get_target_portfolio(self) -> Portfolio:
        positions = []
        nb_assets = len(self.effective_portfolio.positions)
        for position in self.effective_portfolio.positions:
            positions.append(
                position.copy(
                    weighting=Decimal(1 / nb_assets), date=self.trade_date, asset_valuation_date=self.trade_date
                )
            )
        return Portfolio(positions)



from wbcore.metadata.configs.titles import TitleViewConfig


class CustodianTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Custodians"

    def get_instance_title(self):
        return "Custodian"

    def get_create_title(self):
        return "New Custodian"



from wbcore import viewsets
from wbcore.permissions.permissions import InternalUserPermissionMixin

from wbportfolio.models import Rebalancer, RebalancingModel
from wbportfolio.serializers import (
    RebalancerModelSerializer,
    RebalancerRepresentationSerializer,
    RebalancingModelRepresentationSerializer,
)

from ..configs.display import RebalancerDisplayConfig
from ..configs.endpoints import RebalancerEndpointConfig


class RebalancingModelRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:rebalancingmodel"
    queryset = RebalancingModel.objects.all()
    serializer_class = RebalancingModelRepresentationSerializer


class RebalancerRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:rebalancingmodel"
    queryset = Rebalancer.objects.all()
    serializer_class = RebalancerRepresentationSerializer


class RebalancerModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    queryset = Rebalancer.objects.select_related("rebalancing_model", "portfolio")
    serializer_class = RebalancerModelSerializer
    display_config_class = RebalancerDisplayConfig
    endpoint_config_class = RebalancerEndpointConfig



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class PortfolioRoleInstrumentEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:instrument-portfoliorole-list", args=[self.view.kwargs["instrument_id"]], request=self.request
        )



from wbcore import serializers

from wbportfolio.models.portfolio_swing_pricings import PortfolioSwingPricing
from wbportfolio.serializers.portfolios import PortfolioRepresentationSerializer


class PortfolioSwingPricingModelSerializer(serializers.ModelSerializer):
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")

    class Meta:
        model = PortfolioSwingPricing
        fields = (
            "id",
            "portfolio",
            "_portfolio",
            "valid_date",
            "negative_threshold",
            "negative_swing_factor",
            "positive_threshold",
            "positive_threshold",
        )



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class ProductCustomerEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None


class ProductPerformanceFeesEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_instance_endpoint(self, **kwargs):
        return reverse("wbportfolio:product-list", request=self.request)


class InstrumentPriceAUMDataEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbportfolio:aumchart-list", request=self.request)


# Product Portfolio Viewsets
class NominalProductEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbportfolio:product-nominalchart-list", [self.view.kwargs["product_id"]], request=self.request)


class AUMProductEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbportfolio:product-aumchart-list", [self.view.kwargs["product_id"]], request=self.request)
